using System;
using Jypeli;

namespace Cat_Rescue;

/// @author Tomi
/// @version 14.03.2025
/// <summary>
/// 
/// </summary>
public class Cat_Rescue : Game
{
    // Kenttä merkkijonona
    private static readonly String[] level1map = {
        " YYYYYYYYYYYYYYYYYYYYYYYY ",  
        " Y       X              Y ",  
        " Y     0 X           6  Y ",  
        " Y *     X              Y ",  
        " Y       X   XXX  XXX   Y ",  
        " Y       X   X      X   Y ",
        " YXX  XXXX   X      X   Y ",
        " Y           X      X   Y ",
        " Y  5        X      X   Y ",
        " Y           X      X   Y ",
        " YXX  XXXXXXXX      X   Y ",
        " Y       X          X   Y ",
        " Y       X          X   Y ",
        " Y     1 X 2        X   Y ",  
        " Y       X          X   Y ",  
        " Y       X          X C Y ",  
        " YYYYYYYYYYYYYYYYYYYYYYYY ",  
    };
    
    // Kenttän oliotiedot taulukoina
    private int[][] catRoutes = new int[][] {[4, 2, 3, 6, 2, 6, 1, 6, 2, 11, 3, 13, 0, 0], [4, 2, 1, 5, 2, 6, 1, 6, 2, 11, 3, 13, 0, 0],[2, 5, 1, 11, 2, 6, 3, 11, 3, 1, 3, 1, 0, 0]};
    private int[][] guardRoutes = new int[][] {[2, 7, 4, 7, 0, 0], [4, 10, 2, 10, 0, 0]};
    private int catCount = 3;
    private int guardCount = 2;
    private int[] catPos = new int[6];
    private int[] guardPos = new int[4];
    
    // Yksittäisiä pelaajan liikkumiseen liittyviä muuttujia
    private int[] playerPos = new int[2];
    private bool playerMoveCooldown;
    
    // Olioiden kuvat
    private Image catImage = LoadImage("arrowleft.jpg");
    private Image guardImage = LoadImage("arrowleft.jpg");
    private Image lightImage = LoadImage("flashlight.png");
    private Image playerImage = LoadImage("greenArrow.png");
    
    // Kartan luomiseen ja käyttämiseen liittyviä muuttujia
    private int[,] mapEntities = new int[24,24];
    private GameObject player;
    private GameObject key;
    private GameObject[] guard = new GameObject[2];
    private GameObject[] flashlight = new GameObject[2];
    private GameObject[] cage = new GameObject[3];
    private GameObject[] cat = new GameObject[3];
    private GameObject[] safetyOverride = new GameObject[3];
    
    // Pelin tilaan liittyviä muuttujia
    private int score;
    private bool keyHeld;
    private bool gameInProgress;
    
    
    /// <summary>
    /// Ohjelman emsimäinen aliohjelma, jolla luodaan ja käytetään alkuvalikkoa.
    /// Valikkoon kuuluu päävalikko ja kenttävalikko.
    /// </summary>
    public override void Begin()
    {
        // Päävalikko
        string[] options = { "Levels", "Exit" };
        MultiSelectWindow startMenu = new MultiSelectWindow("Main Menu", options);
        Add(startMenu);
        
        // Kenttävalikko
        string[] levels = { "Level 1", "Level 2", "Level 3" , "Back"};
        MultiSelectWindow levelMenu = new MultiSelectWindow("Levels", levels);
        
        // Päävalikkon painikkeiden toiminta
        startMenu.AddItemHandler(0, delegate {Add(levelMenu);});
        startMenu.AddItemHandler(1, Exit);
        
        // Kenttävalikkon painikkeiden toiminta
        levelMenu.AddItemHandler(0, delegate {StartGame();});
        levelMenu.AddItemHandler(1, delegate {Add(startMenu);});
        levelMenu.AddItemHandler(2, delegate {Add(startMenu);});
        levelMenu.AddItemHandler(3, delegate {Add(startMenu);});
           
        Add(startMenu);
        Keyboard.Listen(Key.Escape, ButtonState.Pressed, ConfirmExit, "Lopeta peli");
    }
    

    /// <summary>
    /// Aliohjema aloittaa itse pelin monen eri aliohjelman avulla.
    /// Alotettaviin asioihin kuuluu kartan luonti, pelaajan ohjaus ja vartioiden tekoäly.
    /// </summary>
    private void StartGame()
    {
        // Luo kartan
        CreateMap();
        CreateFloor(new Vector(100, 20), 160, 200, Color.LightGray, 1, 0);
        CreateFloor(new Vector(-200, 150), 160, 200, Color.LightGray, 1, 1);
        
        // Käynnistää pelaajan ohjauksen
        CreateKeys();
        
        // Käynnistää olioiden tekoälyn
        GuardMovement();
        GuardDetection();
        
        gameInProgress = true;
    }
    
    
    /// <summary>
    /// Aliohjema luo kartan merkkijonosta käyttäen Jypelin kirjaston valmiita aliohjelmia.
    /// </summary>
    private void CreateMap()
    {
        // Taustan luominen
        Level.Background.CreateGradient(Color.LightGreen, Color.LightYellow);
        
        // Lattian luominen
        CreateFloor(new Vector(0, 0), 750, 500, Color.LightGray, 0, -1);
        
        // Luo kartan seinät, sekä kaikki peli interaktiiviset oliot 
        TileMap tiles = TileMap.FromStringArray(level1map);
        for (int i = 5; i < 7; i++)
        {
            tiles.SetTileMethod(Convert.ToChar((char)(i+48)), CreateGuard, Color.Salmon, i);
        }
        tiles.SetTileMethod('X', CreateWall, Color.Gray);
        tiles.SetTileMethod('Y', CreateWall, Color.DarkGray);
        tiles.SetTileMethod('*', CreateKey, Color.LightYellow);
        tiles.SetTileMethod('C', CreatePlayer, Color.White);
        for (int i = 0; i < 3; i++)
        {
            tiles.SetTileMethod(Convert.ToChar((char)(i+48)), CreateCageAndCat, Color.LightBlue, i);
        }
        tiles.Execute(32, 32);
        
        Camera.ZoomToLevel();
    }
    
    
    /// <summary>
    /// Aliohjema luo kentän seinät ja tallentaa ne myöhempää käyttöä varten. 
    /// </summary>
    /// <param name="location">Seinän paikka kentässä</param>
    /// <param name="width">Seinän leveys</param>
    /// <param name="height">Seinän korkeus</param>
    /// <param name="color">Seinän väri</param>
    private void CreateWall(Vector location, double width, double height, Color color)
    {
        // Luo seinän
        GameObject wall = new GameObject(width, height);
        wall.Position = location;
        wall.Color = color;
        Add(wall, 2);
        
        // Tallennetaan seinä koordinaatistoon, helppoa viittaamista varten
        mapEntities[Convert.ToInt16(wall.X/32+11.5), Convert.ToInt16(wall.Y/32+8)] = 1;
    }
    
    
    /// <summary>
    /// Aliohjema luo kentän lattian.
    /// Lisäksi se luo ns. piilolattiat, joita tarvitaan vartioiden toimintaa varten. 
    /// </summary>
    /// <param name="location">Lattian paikka kentässä</param>
    /// <param name="width">Lattian leveys</param>
    /// <param name="height">Lattian korkeus</param>
    /// <param name="color">Lattian väri</param>
    /// <param name="layer">Lattian kerros</param>
    /// <param name="safetyBlock">Piilolattian tunnus</param>
    private void CreateFloor(Vector location, double width, double height, Color color, int layer, int safetyBlock) //Aliohjelma, joka luo kentän lattian
    {
        // Luo lattian
        GameObject floor = new GameObject(width, height);
        floor.Position = location;
        floor.Color = color;
        Add(floor, layer);
        
        // Tallentaa piilolattian helppoa viittaamista varten
        if (safetyBlock >= 0)
        {
            safetyOverride[safetyBlock] = floor;
        }
    }
    
    
    /// <summary>
    /// Aliohjema luo kentään häkit ja häkissä olevat kissat, sekä tallentaa ne koordinaatistoon.
    /// </summary>
    /// <param name="location">Häkin(ja kissan) paikka kentässä</param>
    /// <param name="width">Häkin(ja kissan) leveys</param>
    /// <param name="height">Häkin(ja kissan) korkeus</param>
    /// <param name="color">Häkin(ja kissan) väri</param>
    /// <param name="id">Häkin(ja kissan) tunnus</param>
    private void CreateCageAndCat(Vector location, double width, double height, Color color, int id)
    {
        // Luo häkin
        cage[id] = new GameObject(width, height);
        cat[id] = new GameObject(width, height);
        cage[id].Position = location;
        cage[id].Color = color;
        
        // Luo kissan
        cat[id] = new GameObject(width, height);
        cat[id].Position = location;
        cat[id].Color = Color.Purple;
        cat[id].Image = catImage;
        
        // Tallentaa kissan paikkatiedot
        catPos[2*id] = Convert.ToInt16(cat[id].X/32+11.5);
        catPos[2*id+1] = Convert.ToInt16(cat[id].Y/32+8);
        
        //Tallentaa häkin koordinaatistoon, helppoa viittaamista varten
        mapEntities[Convert.ToInt16(cage[id].X/32+11.5), Convert.ToInt16(cage[id].Y/32+8)] = (id + 3);
        
        Add(cat[id], 1);
        Add(cage[id], 2);
    }
    
    
    /// <summary>
    /// Aliohjema luo kentän avaimen ja tallentaa sen koordinaatistoon.
    /// </summary>
    /// <param name="location">Avaimen paikka kentässä</param>
    /// <param name="width">Avaimen leveys</param>
    /// <param name="height">Avaimen korkeus</param>
    /// <param name="color">Avaimen väri</param>
    private void CreateKey(Vector location, double width, double height, Color color) //Aliohjelma, joka luo avaimen kenttään
    {
        // Luo avaimen
        key = new GameObject(width, height);
        key.Position = location;
        key.Color = color;
        Add(key);
       
        //Tallentaa avaimen koordinaatistoon, helppoa viittaamista varten
        mapEntities[Convert.ToInt16(key.X/32+11.5), Convert.ToInt16(key.Y/32+8)] = 2;
    }

    
    /// <summary>
    /// Aliohjema luo kentään vartijan, sekä tallentta sen paikkatiedot.
    /// Lisäksi luo vartijalle taskulampun, jolla pelaaja voidaan ottaa kiinni.
    /// </summary>
    /// <param name="location">Vartijan paikka kentässä</param>
    /// <param name="width">Vartijan leveys</param>
    /// <param name="height">Vartijan korkeus</param>
    /// <param name="color">Vartijan väri</param>
    /// <param name="id">Vartijan tunnus</param>
    private void CreateGuard(Vector location, double width, double height, Color color, int id)
    {
        // Luo vartijan
        guard[id-5] = new GameObject(width, height);
        guard[id-5].Position = location;
        guard[id-5].Color = color;
        guard[id-5].Image = guardImage;
        
        // Tallentaa vartijan paikkatiedot
        guardPos[2*(id-5)] = Convert.ToInt16(guard[id-5].X/32+11.5);
        guardPos[2*(id-5)+1] = Convert.ToInt16(guard[id-5].Y/32+8);
        
        Add(guard[id-5], 0);
        
        // Luo vartijan taskulampun
        flashlight[id-5] = new GameObject(width+90, height+96);
        flashlight[id-5].Position = location - new Vector(80.0, 0.0);
        flashlight[id-5].Color = Color.LightYellow;
        flashlight[id-5].Angle = Angle.FromRadians(Math.PI*3/2);
        flashlight[id-5].Image = lightImage;
        
        guard[id-5].Add(flashlight[id-5]);
    }
    
    
    /// <summary>
    /// Aliohjema luo pelaajan ja tallentaa sen paikkatiedot.
    /// </summary>
    /// <param name="location">Pelaajan paikka kentässä</param>
    /// <param name="width">Pelaajan leveys</param>
    /// <param name="height">Pelaajan korkeus</param>
    /// <param name="color">Pelaajan väri</param>
    private void CreatePlayer(Vector location, double width, double height, Color color) //Aliohjelma, joka luo pelaaja hahmon ja määrää pelaajan muuttujat
    {
        // Luo pelaajan
        player = new GameObject(width, height);
        player.Position = location;
        player.Color = color;
        player.Image = playerImage;
        Add(player, 2);
        
        // Tallentaa pelaajan paikkatiedot
        playerPos[0] = 21;
        playerPos[1] = 1;
    }
    
    
    /// <summary>
    /// Aliohjema aloittaa näppäinten kuuntelun ja näin kääntää näppäinten painallukset pelaajan liikkeeeksi.
    /// </summary>
    private void CreateKeys()
    {
        Keyboard.Listen(Key.W, ButtonState.Pressed, MovePlayer, "Liikkuu ylös", 0, 1, Math.PI/2);
        Keyboard.Listen(Key.A, ButtonState.Pressed, MovePlayer, "Liikkuu vasemmalle", -1, 0, Math.PI);
        Keyboard.Listen(Key.S, ButtonState.Pressed, MovePlayer, "Liikkuu alas", 0, -1, Math.PI*3/2);
        Keyboard.Listen(Key.D, ButtonState.Pressed, MovePlayer, "Liikkuu oikealle", 1, 0, 0.0);
    }
    
    
    /// <summary>
    /// Aliohjema, joka liikutta pelaajaa ja aloitaa kaiken vuorovaikutuksen kentän kanssa.
    /// </summary>
    /// <param name="xMove">Liike x-akselin suunnassa</param>
    /// <param name="yMove">Liike y-akselin suunnassa</param>
    /// <param name="angle">Liikkeen suunnan kulma</param>
    private void MovePlayer(int xMove, int yMove, double angle)
    {
        // Estää liikkuumisen, kon edellinen liike on kesken
        if (!playerMoveCooldown)
        {
            // Kääntää pelaajahahmon liikkeen suuntaan
            player.Angle = Angle.FromRadians(angle);
            
            // Estää liikkumisen, jos edessä on seinä
            if (mapEntities[playerPos[0] + xMove, playerPos[1] + yMove] != 1)
            {
                // Liikuttaa pelaajan
                playerMoveCooldown = true;
                player.MoveTo(new Vector((playerPos[0]-11.5+ xMove)*32, (playerPos[1]-8+yMove)*32), 80, PlayerPosSet(xMove, yMove, 80));
       
                // Tarkistaa, jos edessä on avain ja kerää(+ tuhoaa) avaimen mikäli tämä on tosi
                if (mapEntities[playerPos[0], playerPos[1]] == 2)
                {
                    key.Destroy();
                    keyHeld = true;
                    mapEntities[playerPos[0], playerPos[1]] = 0;
                }

                // Tarkistaa, jos edessä on häkki (ja onko pelaajalla avain)
                if (mapEntities[playerPos[0], playerPos[1]] > 2 && keyHeld)
                {
                    for (int i = 0; i < 4; i++)
                    {
                        if (mapEntities[playerPos[0], playerPos[1]] == (i + 3))
                        {
                            // Tuhoaa häkin ja aloitaan kissan liikkeen
                            cage[i].Destroy();
                            ObjectMove(cat[i], catRoutes[i], catPos[2*i], catPos[2*i+1], 6,0, 100, true);
                            mapEntities[playerPos[0], playerPos[1]] = 0;
                        }
                    }
                }
            }
        }
    }


    /// <summary>
    /// Aliohjema, joka aloittaa vartioiden liikkeen.
    /// </summary>
    private void GuardMovement()
    {
        for (int i = 0; i < 2; i++)
        {
            ObjectMove(guard[i], guardRoutes[i], guardPos[2*i], guardPos[2*i+1], 0, 2, 50, false);
        }
    }
    
    
    /// <summary>
    /// Aliohjema, joka ajastaa pelin olioiden (ei pelaaja) liikkeet taulukon avulla.
    /// Lisäksi antaa pisteistä kissoja karatessa ja lopettaa, kun kaikki kissat ovat karanneet
    /// </summary>
    /// <param name="entity">Olio, jota liikutetaan</param>
    /// <param name="route">Taulukko, jossa olion liiketiedot</param>
    /// <param name="posX">Olion x-koordinaatti</param>
    /// <param name="posY">Olion y-koordinaatti</param>
    /// <param name="cycles">Liikkeen jaksojen määrä (0 = ikuinen liike)</param>
    /// <param name="delay">Jaksojen väliset tauot</param>
    /// <param name="speed">Liikkeen nopeus</param>
    /// <param name="destroy">Tuhotaanko olio liikkeen lopuksi</param>
    private void ObjectMove(GameObject entity, int[] route, int posX, int posY, int cycles, int delay, int speed, bool destroy)
    {
        // Ajastin, jolla liike ajastetaan
        Timer timer = new Timer(0.1);
        timer.Timeout += delegate
        {
            // Aliohjelma, joka liikuttaa oliota
            MovementFunction(entity, route, posX, posY, speed);
            
            // Tallentaa olion uudet paikkatiedot
            if (route[2 * (route[route.Length - 1])] == 1) //ylös
            {
                posY += route[2 * (route[route.Length - 1]) + 1];
            }
            if (route[2 * (route[route.Length - 1])] == 2) //oikealle
            {
                posX += route[2 * (route[route.Length - 1]) + 1];
            }
            if (route[2 * (route[route.Length - 1])] == 3) //alas
            {
                posY += -route[2 * (route[route.Length - 1]) + 1];
            }
            if (route[2 * (route[route.Length - 1])] == 4) //vasemmalle
            {
                posX += -route[2 * (route[route.Length - 1]) + 1];
            }
            route[route.Length-1] += 1;
            
            // Ajastimen kestoksi nykyisen liikkeen kesto + tauon kesto
            timer.Interval = ((route[2*(route[route.Length-1])-1]*32.0/speed) + delay);
            
            // Ikuisen liikkeen jatkaminen
            if (route[route.Length-1] == 2  && cycles == 0)
            {
                route[route.Length-1] = 0;
            }
            
            // Rajallisen liikkeen lopetus
            if (route[route.Length-1] == 7  && destroy)
            {
                Remove(entity);
                
                // Pisteiden päivitys ja voiton tarkastus
                score++;
                if (score == catCount)
                {
                    // Peli loppuu voittoon
                    GameOver(true);
                }
            }
        };
        
        // Ikuisen liikkeen aloitus
        if (cycles == 0)
        {
            timer.Start();
        }
        // Rajallisen liikkeen aloitus
        else
        {
            timer.Start(cycles+1);
        }
    }
    
    
    /// <summary>
    /// Aliohjema, joka liikuttaa ja kääntää pelin olioita (ei pelaaja), taulukon avulla.
    /// </summary>
    /// <param name="entity">Olio, jota liikutetaan</param>
    /// <param name="route">Taulukko, jossa olion liiketiedot</param>
    /// <param name="posX">Olion x-koordinaatti</param>
    /// <param name="posY">Olion y-koordinaatti</param>
    /// <param name="speed">Liikkeen nopeus</param>
    /// <returns>Jypelin aliohjelman vaatima palautus (palauttaa null)</returns>
    private void MovementFunction(GameObject entity, int[] route, int posX, int posY, int speed)
    {
        if (route[2*(route[route.Length-1])] == 1) //ylös
        {
            entity.MoveTo(new Vector((posX - 11.5) * 32, (route[2*(route[route.Length-1])+1] + posY - 8) * 32), speed);
            entity.Angle = Angle.FromRadians(Math.PI*3/2);
        }
        if (route[2*(route[route.Length-1])] == 2) //oikealle
        {
            entity.MoveTo(new Vector((route[2*(route[route.Length-1])+1] + posX - 11.5) * 32, (posY - 8) * 32), speed);
            entity.Angle = Angle.FromRadians(Math.PI);
        }
        if (route[2*(route[route.Length-1])] == 3) //alas
        {
            entity.MoveTo(new Vector((posX - 11.5) * 32, (- route[2*(route[route.Length-1])+1] + posY - 8) * 32), speed);
            entity.Angle = Angle.FromRadians(Math.PI/2);
        }
        if (route[2 * (route[route.Length - 1])] == 4) //vasemmalle
        {
            entity.MoveTo(new Vector((-route[2 * (route[route.Length - 1]) + 1] + posX - 11.5) * 32, (posY - 8) * 32), speed);
            entity.Angle = Angle.FromRadians(0);
        }
    }
    
    
    /// <summary>
    /// Aliohjema tallentaa pelaajan sijainnin koordinaatiostoon joka kerta, kun pelaaja liikkuu.
    /// Lisäksi asettaa aikarajoituksen pelaajan liikkumiselle.
    /// </summary>
    /// <param name="xMove">Liike x-akselin suunnassa</param>
    /// <param name="yMove">Liike y-akselin suunnassa</param>
    /// <param name="speed">Liikkeen nopeus</param>
    /// <returns>Jypelin aliohjelman vaatima palautus (palauttaa null)</returns>
    private Action PlayerPosSet(int xMove, int yMove, int speed)
    {
        // Päivittää pelaajan koordinaatit
        playerPos[0] += xMove;
        playerPos[1] += yMove;
        
        // Ajastin, joka rajoitaa pelaajan liikettä
        Timer movecooldown = new Timer();
        movecooldown.Interval = 30.0/speed; //movement cooldown!
        movecooldown.Timeout += delegate
        {
            playerMoveCooldown = false;
        };
        movecooldown.Start(1);
        
        return null;
    }

    
    /// <summary>
    /// Aliohjema toimii osana vartijoiden tekoälyä, jolla pelaaja voidaan ottaa kiiinni.
    /// </summary>
    private void GuardDetection()
    {
        // Ajastin kiinniottotarkistukselle
        Timer flashDetection = new Timer(0.5);
        flashDetection.Timeout += delegate
        {
            // Tarkistaa kaikkien vartjoiden lamput
            for (int i = 0; i < guardCount; i++)
            {
                // Tarkistaa onko pelaaja lampunvalaisemalla alueella
                if (flashlight[i].IsInside(player.Position))
                {
                    // Tarkistaa onko pelaaja piilolattian päällä (vartijan ja pelaaja välissä seinä)
                    if (!(safetyOverride[0].IsInside(player.Position)) && !(safetyOverride[1].IsInside(player.Position)))
                    {
                        // Peli loppuu häviöön
                        GameOver(false);
                    }

                }
            }
        }; 
        // Käynnistää ajastimen
        flashDetection.Start();
    }

    
    /// <summary>
    /// Aliohjema lopetaa pelin joko voittoon tai häviöön.
    /// </summary>
    /// <param name="win">Voitettiinko peli (tosi=voitto, epätosi=häviö)</param>
    private void GameOver(bool win)
    {
        if (gameInProgress)
        {
            // Lopettaa pelin
            gameInProgress = false;
            ClearGameObjects();
            ClearTimers();
            
            // Voittoviesti
            if (win)
            {
                string[] gameOverOptions = {"Back to Menu"};
                MultiSelectWindow gameOverMenu = new MultiSelectWindow("Game Won!", gameOverOptions);
                Add(gameOverMenu);
            
            
                gameOverMenu.AddItemHandler(0, delegate
                {
                    Begin();
                }); 
            }
            
            // Häviöviesti
            else
            {
                string[] gameOverOptions = {"Back to Menu"};
                MultiSelectWindow gameOverMenu = new MultiSelectWindow("Caught by a Guard!", gameOverOptions);
                Add(gameOverMenu);
                
                gameOverMenu.AddItemHandler(0, delegate
                {
                    Begin();
                }); 
            }
        }
    }
}